# -*- coding: utf-8 -*-

import os
import sys

import numpy as np
from matplotlib.backends.backend_qtagg  import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from qtpy import uic
from qtpy.QtWidgets import (QApplication, QFileDialog, QMainWindow, QSizePolicy, QVBoxLayout,QTextEdit)
from lmfit import Parameters, Minimizer, fit_report, report_fit


from ..energy import Energy
from ..instrument.tools import get_tau, _cleanargs, _star, _modvec
from ..instrument.tas_spectr import TripleAxisSpectr
from ..insfit import FitConv
from .tools import SqwQScanTwoPeaks, PrefDemoFF, angle2, SelFormFactor



class MyMplCanvas(FigureCanvas):
    def __init__(self, parent=None, width=261, height=201, dpi=100, qslice='QxQy'):
        self.fig = Figure(figsize=(width, height), dpi=dpi, edgecolor='k')
        self.fig.patch.set_facecolor('#F0F0F0')
        self.fig.subplots_adjust(bottom=0.25, left=0.25)

        self.axes = self.fig.add_subplot(111)
        self.axes.set_position([0.25,0.25,0.73,0.73])

        #self.compute_initial_figure(self.axes, qslice, projections, u, v)

        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

    def compute_initial_figure(self, qslice, projections, u, v):
        pass


class MyDataCanvas(FigureCanvas):
    def __init__(self, parent=None, width=261, height=201, dpi=300):
        self.fig = Figure(figsize=(width, height), dpi=dpi, edgecolor='k')
        self.fig.patch.set_facecolor('#F0F0F0')
        self.fig.subplots_adjust(bottom=0.25, left=0.25)
        

        self.axes = self.fig.add_subplot(111)
        self.axes.set_position([0.17,0.17,0.81,0.81])
        self.axes.set_xlabel('Energy [meV]', fontsize=10)
        self.axes.set_ylabel('Intensity [a.u]', fontsize=10)

        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)



class MyStaticMplCanvas(MyMplCanvas):
    def __init__(self, *args, **kwargs):
        super(MyStaticMplCanvas, self).__init__(*args, **kwargs)

    def compute_initial_figure(self, axis, qslice, projections, u, v):
        self.plot_slice(axis, qslice, projections, u, v)


class MainWindow(QMainWindow):
    r"""Main Window of Resolution Calculator

    """
    def closeEvent(self, event):
        QApplication.quit()
        
    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)

        uic.loadUi(os.path.join(os.path.dirname(__file__), 'ui','ResConFitQScanFF.ui'), self)

        self.qxqyplot = QVBoxLayout(self.qx_qy_plot_widget)
        self.qxwplot  = QVBoxLayout(self.qx_w_plot_widget)
        self.qywplot  = QVBoxLayout(self.qy_w_plot_widget)
        self.dataplot = QVBoxLayout(self.fit_data_plot)

        self.text_output.setFontPointSize(6)
        self.text_output.setLineWrapMode(QTextEdit.NoWrap)

        self.dir_dict    = {'Clockwise': 1, 'Counter-Clockwise': -1}
        self.infin_dict  = {'ki': 1, 'kf': -1}

        self.method_dict = {'Cooper-Nathans': 0, 'Popovici': 1}
        self.moncor_dict = {'On': 1, 'Off': 0}
        
        self.param       = np.array([1.0,1, 1, 1, 1, 1, 1, 1])
        self.param_fixed = np.array([1,  1, 1, 1, 1, 1, 1, 0])
        self.data        = []
        self.data_start  = 0
        self.data_end    = 0
        
        self.dplot=None

        self.load_instrument()
        self.load_signals()
        

    def load_instrument(self):
        self.param       = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        self.param_fixed = np.array([1,   1, 1, 1, 1, 1, 1, 0])

        self.instrument = TripleAxisSpectr(efixed=5)

        self.instrument.sample.a, self.instrument.sample.b, self.instrument.sample.c = [float(i) for i in
                                                                                        self.abc_input.text().split(',')]
        self.instrument.sample.alpha, self.instrument.sample.beta, self.instrument.sample.gamma = [float(i) for i in
                                                                                                   self.abg_input.text().split(',')]
        #Deng: the following statement has been moved from __init__ to here, in order to run the calculation after modifying energy
        self.edrop_dict = {'energy (meV)': float(self.energy_input.text()),
                           'wavelength (A)': Energy(wavelength=float(self.energy_input.text())).energy,
                           'wave vector (A-1)': Energy(wavevector=float(self.energy_input.text())).energy}
        self.instrument.sample.height     =    float(self.sample_height_input.text())
        self.instrument.sample.width      =    float(self.sample_width_input.text())
        self.instrument.sample.depth      =    float(self.sample_depth_input.text())
        self.instrument.sample.u          =   [float(i) for i in self.sample_u_input.text().split(',')]
        self.instrument.sample.v          =   [float(i) for i in self.sample_v_input.text().split(',')]
        self.instrument.sample.dir        =   self.dir_dict[self.sample_dir_select.currentText()]
        self.instrument.sample.mosaic     =    float(self.sample_mosaic_input.text())
        self.instrument.sample.vmosaic    =    float(self.sample_vmosaic_input.text())
        self.instrument.sample.shape_type =   self.sample_shape_dropdown.currentText().lower()
        
        if self.mono_select_dropdown.currentText() == 'Custom':
            self.instrument.mono.tau = 2 * np.pi / float(self.mono_select_input)
        else:
            self.instrument.mono.tau = get_tau(self.mono_select_dropdown.currentText())
            self.mono_select_input.setText(
                '{0:.3f}'.format(2. * np.pi / get_tau(self.mono_select_dropdown.currentText())))
        self.instrument.mono.mosaic  = float(self.mono_mosaic_input.text())
        self.instrument.mono.vmosaic = float(self.mono_vmosaic_input.text())
        self.instrument.mono.dir     = self.dir_dict[self.mono_dir_select.currentText()]
        self.instrument.mono.height  = float(self.mono_height_input.text())
        self.instrument.mono.width   = float(self.mono_width_input.text())
        self.instrument.mono.depth   = float(self.mono_depth_input.text())

        if self.ana_select_dropdown.currentText() == 'Custom':
            self.instrument.ana.tau  = 2 * np.pi / float(self.ana_select_input)
        else:
            self.instrument.ana.tau  = get_tau(self.ana_select_dropdown.currentText())
            self.ana_select_input.setText('{0:.3f}'.format(2. * np.pi / get_tau(self.ana_select_dropdown.currentText())))
        self.instrument.ana.mosaic   = float(self.ana_mosaic_input.text())
        self.instrument.ana.vmosaic  = float(self.ana_vmosaic_input.text())
        self.instrument.ana.dir      = self.dir_dict[self.ana_dir_select.currentText()]
        self.instrument.ana.height   = float(self.ana_height_input.text())
        self.instrument.ana.width    = float(self.ana_width_input.text())
        self.instrument.ana.depth    = float(self.ana_depth_input.text())

        self.instrument.efixed       = self.edrop_dict[self.energy_dropdown.currentText()]

        self.instrument.infin        = self.infin_dict[self.fixed_kikf_dropdown.currentText()]
        self.instrument.hcol         = [float(i) for i in self.hcols_input.text().split(',')]
        self.instrument.vcol         = [float(i) for i in self.vcols_input.text().split(',')]
        self.instrument.arms         = [float(i) for i in self.arms_input.text().split(',')]

        self.instrument.guide.height = float(self.guide_height_input.text())
        self.instrument.guide.width  = float(self.guide_width_input.text())

        self.instrument.detector.height = float(self.detector_height_input.text())
        self.instrument.detector.width  = float(self.detector_width_input.text())

        self.instrument.moncor       = self.moncor_dict[self.moncor_dropdown.currentText()]
        self.instrument.method       = self.method_dict[self.method_dropdown.currentText()]

        if self.mono_hcurve_input.text() != 'None':
            self.instrument.mono.rh       = float(self.mono_hcurve_input.text())
        if self.mono_vcurve_input.text() != 'None':
            self.instrument.mono.rv       = float(self.mono_vcurve_input.text())

        if self.ana_hcurve_input.text()  != 'None':
            self.instrument.ana.rh        = float(self.ana_hcurve_input.text())
        if self.ana_vcurve_input.text()  != 'None':
            self.instrument.ana.rv        = float(self.ana_vcurve_input.text())
        self.instrument.description_string = ''

        self.q = [float(i) for i in self.q_input.text().split(',')]
        self.w = [float(i) for i in self.w_input.text().split(',')]
        
        
        self.mag_on =self.chkMagFF.isChecked()
        if self.mag_on:
            self.mag_ion.setEnabled(True)
            self.mag_form_factor.setEnabled(True)
            self.mag_ion_name = self.mag_ion.text()
            self.ffactor = SelFormFactor(self.mag_ion_name)
            if self.ffactor is None:
                self.ffactor = SelFormFactor("Mn2")
                print("The given magnetic ion was not found. Instead, Mn2 is used.")

            AA=self.ffactor["AA"]
            aa=self.ffactor["aa"]
            BB=self.ffactor["BB"]
            bb=self.ffactor["bb"]
            CC=self.ffactor["CC"]
            cc=self.ffactor["cc"]
            DD=self.ffactor["DD"]

            self.mag_form_factor.setText( "{}  {}  {}  {}  {}  {}  {}".format(AA, aa, BB, bb, CC, cc, DD))
        else:
            self.mag_ion_name = "NONE"
            self.ffactor = None
            self.mag_ion.setEnabled(False)
            self.mag_form_factor.setText( "------------------------")
            self.mag_form_factor.setEnabled(False)
            print("No magnetic form factor is set.")

        [length, temph,tempk, templ, tempW] = _cleanargs(self.q[0],self.q[1],self.q[2],self.w)
        self.hkle  =  [temph, tempk, templ, tempW]

        #self.instrument.calc_resolution(self.q)
        #self.instrument.calc_projections(self.q)
        
        self.param[0] = float(self.fit_param_p1.text())
        self.param[1] = float(self.fit_param_p2.text())
        self.param[2] = float(self.fit_param_ratio.text())
        self.param[3] = float(self.fit_param_w1.text())
        self.param[4] = float(self.fit_param_w2.text())
        self.param[5] = float(self.fit_param_int.text())
        self.param[6] = float(self.fit_param_bg.text())
        self.param[7] = float(self.fit_param_temp.text())
        
        
        self.data_start   = int(self.fit_data_start.text())
        self.data_end     = int(self.fit_data_end.text())
        self.str_param_fixed   = self.fit_param_fixed.text()
        self.param_fixed  = [int(ii) for ii in self.str_param_fixed.split() if ii.isdigit()]

        # TEST CODE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        try:
            self.clearLayout(self.qxqyplot)
            self.clearLayout(self.qxwplot)
            self.clearLayout(self.qywplot)
            #self.clearLayout(self.dataplot)
        except:
            pass

        qxqy = MyStaticMplCanvas(self.qx_qy_plot_widget, width=261, height=201, dpi=100, qslice='QxQy')
        self.qxqyplot.addWidget(qxqy)
        self.instrument.ResolutionPlotProj(ax=qxqy.axes, qslice='QxQy',hkle=self.hkle)

        qxw = MyStaticMplCanvas(self.qx_w_plot_widget,   width=261, height=201, dpi=100, qslice='QxW')
        self.qxwplot.addWidget(qxw)
        self.instrument.ResolutionPlotProj(ax=qxw.axes,  qslice='QxE',hkle=self.hkle)

        qyw = MyStaticMplCanvas(self.qy_w_plot_widget,   width=261, height=201, dpi=100, qslice='QyW')
        self.qywplot.addWidget(qyw)
        self.instrument.ResolutionPlotProj(ax=qyw.axes,  qslice='QyE',hkle=self.hkle)
        
        if self.dplot is None:
            self.dplot = MyDataCanvas(self.fit_data_plot,width=261, height=201, dpi=100)
            self.dataplot.addWidget(self.dplot)
        else:
            self.dplot.fig.canvas.draw_idle()

        self.text_output.setText(str(self.instrument.description_string))
        

    def load_signals(self):
        self.method_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.mono_dir_select.currentIndexChanged.connect(self.load_instrument)
        self.sample_dir_select.currentIndexChanged.connect(self.load_instrument)
        self.ana_dir_select.currentIndexChanged.connect(self.load_instrument)
        self.mono_select_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.ana_select_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.moncor_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.fixed_kikf_dropdown.currentIndexChanged.connect(self.load_instrument)

        self.energy_input.editingFinished.connect(self.load_instrument)
        self.mono_select_input.editingFinished.connect(self.load_instrument)
        self.mono_mosaic_input.editingFinished.connect(self.load_instrument)
        self.mono_vmosaic_input.editingFinished.connect(self.load_instrument)
        self.mono_height_input.editingFinished.connect(self.load_instrument)
        self.mono_width_input.editingFinished.connect(self.load_instrument)
        self.mono_depth_input.editingFinished.connect(self.load_instrument)
        self.mono_hcurve_input.editingFinished.connect(self.load_instrument)
        self.mono_vcurve_input.editingFinished.connect(self.load_instrument)

        self.ana_select_input.editingFinished.connect(self.load_instrument)
        self.ana_mosaic_input.editingFinished.connect(self.load_instrument)
        self.ana_vmosaic_input.editingFinished.connect(self.load_instrument)
        self.ana_height_input.editingFinished.connect(self.load_instrument)
        self.ana_width_input.editingFinished.connect(self.load_instrument)
        self.ana_depth_input.editingFinished.connect(self.load_instrument)
        self.ana_hcurve_input.editingFinished.connect(self.load_instrument)
        self.ana_vcurve_input.editingFinished.connect(self.load_instrument)

        self.abc_input.editingFinished.connect(self.load_instrument)
        self.abg_input.editingFinished.connect(self.load_instrument)
        self.sample_mosaic_input.editingFinished.connect(self.load_instrument)
        self.sample_vmosaic_input.editingFinished.connect(self.load_instrument)
        self.sample_height_input.editingFinished.connect(self.load_instrument)
        self.sample_width_input.editingFinished.connect(self.load_instrument)
        self.sample_depth_input.editingFinished.connect(self.load_instrument)
        self.sample_u_input.editingFinished.connect(self.load_instrument)
        self.sample_v_input.editingFinished.connect(self.load_instrument)
        self.sample_shape_dropdown.currentIndexChanged.connect(self.load_instrument)

        self.hcols_input.editingFinished.connect(self.load_instrument)
        self.vcols_input.editingFinished.connect(self.load_instrument)
        self.arms_input.editingFinished.connect(self.load_instrument)

        self.guide_height_input.editingFinished.connect(self.load_instrument)
        self.guide_width_input.editingFinished.connect(self.load_instrument)
        self.detector_height_input.editingFinished.connect(self.load_instrument)
        self.detector_width_input.editingFinished.connect(self.load_instrument)

        self.energy_input.returnPressed.connect(self.load_instrument)
        
        #deng: add the following statement in order to recalc resolution when change q and w
        self.q_input.editingFinished.connect(self.load_instrument)
        self.w_input.editingFinished.connect(self.load_instrument)
        self.fit_param_p1.editingFinished.connect(self.load_instrument)
        self.fit_param_p2.editingFinished.connect(self.load_instrument)
        self.fit_param_ratio.editingFinished.connect(self.load_instrument)
        self.fit_param_w1.editingFinished.connect(self.load_instrument)
        self.fit_param_w2.editingFinished.connect(self.load_instrument)
        self.fit_param_int.editingFinished.connect(self.load_instrument)
        self.fit_param_bg.editingFinished.connect(self.load_instrument)
        self.fit_param_temp.editingFinished.connect(self.load_instrument)
        
        self.fit_param_fixed.editingFinished.connect(self.load_instrument)
        self.fit_data_start.editingFinished.connect(self.load_instrument)
        self.fit_data_end.editingFinished.connect(self.load_instrument)
        self.fit_btn_init.pressed.connect(self.initData)
        self.fit_btn_fit.pressed.connect(self.fitData)
        self.fit_opendatafile.pressed.connect(self.loadData)
        self.chkMagFF.stateChanged.connect(self.load_instrument)
        self.mag_ion.editingFinished.connect(self.load_instrument)
        #self.mag_form_factor.editingFinished.connect(self.load_instrument)

    def clearLayout(self, layout):
        while layout.count():
            child = layout.takeAt(0)
            child.widget().deleteLater()
    
    def loadData(self):
        fname = QFileDialog.getOpenFileName(self, "open file", "", "data file(*.hklw)")
        #print(fname)
        if  not ( fname[0] == '') :
            self.fit_filepath.setText(fname[0])
            with open(fname[0], 'r') as f:
                self.data      = np.loadtxt(f, unpack=True)
            [H, K, L, W, Iobs] = self.data
            dIobs           =   np.sqrt(Iobs)
            self.data       =   np.array([H, K, L, W, Iobs, dIobs])
            self.data_start =   0
            self.data_end   =   self.data.shape[1]
            
            self.fit_data_start.setText(str(self.data_start))
            self.fit_data_end.setText(str(self.data_end))
            self.dplot.axes.clear()
            self.dplot.axes.plot(H,Iobs, "bo")
            self.dplot.axes.set_xlabel('Q [rlu]', fontsize=8)
            self.dplot.axes.set_ylabel('Intensity [a.u]', fontsize=8)
            self.dplot.fig.canvas.draw_idle()
        
    
    def initData(self):
        self.load_instrument()
        [H, K, L, W, Iobs, dIobs] = self.data[:,self.data_start:self.data_end]
        
        if self.chkMagFF.isChecked(): 
            AA=self.ffactor["AA"]
            aa=self.ffactor["aa"]
            BB=self.ffactor["BB"]
            bb=self.ffactor["bb"]
            CC=self.ffactor["CC"]
            cc=self.ffactor["cc"]
            DD=self.ffactor["DD"]
            self.param       = np.append(self.param, np.array([AA, aa, BB, bb, CC, cc, DD]))
            self.param_fixed = np.append(self.param_fixed, np.array([ 0,  0,  0,  0,  0,  0,  0]))


        sim_inital  = self.instrument.ResConv(sqw=SqwQScanTwoPeaks, pref=PrefDemoFF, nargout=2, hkle=[H,K,L,W], METHOD='fix', ACCURACY=None, p=self.param)
        #print(H,K, L, W, sim_inital)
        self.dplot.axes.clear()
        self.dplot.axes.plot(H,Iobs, "bo", H,sim_inital, "g-")
        self.dplot.fig.canvas.draw_idle()
        
        
    
    def fitData(self):
        #fit the data using the input parameters:
        self.load_instrument()
        [H, K, L, W, Iobs, dIobs] = self.data[:,self.data_start:self.data_end]

        if self.chkMagFF.isChecked():
            AA=self.ffactor["AA"]
            aa=self.ffactor["aa"]
            BB=self.ffactor["BB"]
            bb=self.ffactor["bb"]
            CC=self.ffactor["CC"]
            cc=self.ffactor["cc"]
            DD=self.ffactor["DD"]

            self.param       = np.append(self.param, np.array([AA, aa, BB, bb, CC, cc, DD]))
            self.param_fixed = np.append(self.param_fixed, np.array([ 0,  0,  0,  0,  0,  0,  0]))


    
        fitter     =    FitConv(self.instrument,SqwQScanTwoPeaks,PrefDemoFF,[H,K,L,W],Iobs, dIobs)

        [final_param, dpa, chisqN, sim, CN, PQ, nit, kvg, details] = fitter.fitwithconv(self.instrument,SqwQScanTwoPeaks,PrefDemoFF,[H,K,L,W],Iobs, dIobs, param=self.param, paramfixed=self.param_fixed)
        
        #final = self.instrument.ResConv(sqw=SqwDemo, pref=PrefDemo, nargout=2, hkle=[H,K,L,W], METHOD='fix', ACCURACY=None, p=final_param)

        self.dplot.axes.plot(H,sim, "r-")
        self.dplot.fig.canvas.draw_idle()
        str_output="The fitted parameters:\n"
        parlist=["p1     ","p2     ","ratio   ","w1","w2", "intensity","bg      " ]

        #str_output = "The fitted parameters:\n"
        #parlist    = ["p1     ","p2     ","Ratio   ","w1     ","w2     ", "Intensity","bg      ", "temp    " ]
        
        for index, (iname, ipar, ierr) in enumerate(zip(parlist,final_param,dpa)):
            str_output=str_output+"Param{0}({1}):\t {2:8f}\t {3:8f}\n".format(index, iname, ipar, ierr)
            
        str_output+="The fitted data:\n    H \t    K \t  L  \t  W  \t  Iobs \t dIobs\t Ifit\n"
        for index, (iH, iK, iL, iW, iIobs, idIobs, iFit) in enumerate(zip(H,K,L,W,Iobs,dIobs,sim)):
            str_output=str_output+"{0:8f} \t {1:8f} \t {2:8f} \t {3:8f} \t {4:8f} \t {5:8f} \t {6:8f} \n".format(iH, iK, iL, iW, iIobs, idIobs, iFit)
            
        self.fit_output_text.clear()
        self.fit_output_text.insertPlainText(str_output)


    
        
def main():        

    app = QApplication(sys.argv)
    w = MainWindow()
    w.show()
    sys.exit(app.exec_())



if __name__ == "__main__" :
    main()
    


# -*- coding: utf-8 -*-
# This file is adapted from neutronpy

#Sample class for e.g. Instrument class

import numpy as np
from .lattice import Lattice

def gram_schmidt(in_vecs, row_vecs=True, normalize=True):
    # for calculating the normal direction of two vectors
    if not row_vecs:
        in_vecs = in_vecs.T

    out_vecs = in_vecs[0:1, :]

    for i in range(1, in_vecs.shape[0]):
        proj = np.diag((in_vecs[i, :].dot(out_vecs.T) / np.linalg.norm(out_vecs, axis=1) ** 2).flat).dot(out_vecs)
        out_vecs = np.vstack((out_vecs, in_vecs[i, :] - proj.sum(0)))

    if normalize:
        out_vecs = np.diag(1 / np.linalg.norm(out_vecs, axis=1)).dot(out_vecs)

    if not row_vecs:
        out_vecs = out_vecs.T

    return out_vecs


    

class Sample(Lattice):
    #Sample class 

    def __init__(self, a, b, c, alpha, beta, gamma, u=None, v=None, mosaic=None, vmosaic=None, direct=1,
                 width=None, height=None, depth=None, shape='rectangular', distance=None):
        super(Sample, self).__init__(a, b, c, alpha, beta, gamma)
        if u is not None:
            self._u = np.array(u)
        else:
            self._u = np.array([1,0,0])
        if v is not None:
            self._v = np.array(v)
        else:
            self._v = np.array([0,0,1])
            

        if mosaic is not None:
            self.mosaic = mosaic
        if vmosaic is not None:
            self.vmosaic = vmosaic

        self.dir = direct
        self.shape_type = shape

        if width is not None:
            self.width = width
        if height is not None:
            self.height = height
        if depth is not None:
            self.depth = depth
        if distance is not None:
            self.distance = distance

    def __repr__(self):
        args = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in ['a', 'b', 'c', 'alpha', 'beta', 'gamma']])
        kwargs = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in
                            ['u', 'v', 'mosaic', 'vmosaic', 'direct', 'width', 'height', 'depth', 'shape'] if
                            getattr(self, key, None) is not None])
        return "Sample({0}, {1})".format(args, kwargs)

    def __eq__(self, right):
        self_parent_keys = sorted(list(self.__dict__.keys()))
        right_parent_keys = sorted(list(right.__dict__.keys()))

        if not np.all(self_parent_keys == right_parent_keys):
            return False

        for key, value in self.__dict__.items():
            right_parent_val = getattr(right, key)
            if not np.all(value == right_parent_val):
                print(value, right_parent_val)
                return False

        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    @property
    def u(self):
        r"""First orientation vector
        """
        return self._u

    @u.setter
    def u(self, vec):
        self._u = np.array(vec)

    @property
    def v(self):
        r"""Second orientation vector
        """
        return self._v

    @v.setter
    def v(self, vec):
        self._v = np.array(vec)

    @property
    def direct(self):
        return self._dir

    @direct.setter
    def direct(self, value):
        self._dir = value

    @property
    def Umatrix(self):
        r"""Rotation matrix that rotates the sample's reference frame into the spectrometer's
        """
        ortho_basis = gram_schmidt(np.vstack((self.u, self.v)))

        return np.vstack((ortho_basis, np.cross(ortho_basis[0], ortho_basis[1])))

    @property
    def UBmatrix(self):
        r"""Orientation matrix of the sample
        """
        return self.Umatrix * self.Bmatrix

    def get_phi(self, Q):
        u"""Get out-of-plane scattering angle.
        Parameters
        ----------
        hkl: array_like
        wavelength : float
        Returns
        -------
        phi : float
            The out-of-plane angle
        """
        return self.get_angle_between_planes(Q, np.cross(self.u, self.v))



# -*- coding: utf-8 -*-
# This file is adapted from neutronpy

"""
Useful constants for neutron scattering calculations, including:

* ``magnetic_form_factors()`` : Magnetic Ion j-values
* ``periodic_table()`` :        Periodic table values
* ``scattering_lengths()`` :    Neutron scattering lengths
* ``symmetry()`` :              Space group information
* ``JOULES_TO_MEV`` :           Joules-to-meV conversion factor
* ``BOLTZMANN_IN_MEV_K`` :      Boltzmann constant in meV/K
* ``N_A`` :                     Avogadro constant
* ``neutron_mass`` :            Mass of a neutron in grams
* ``e`` :                       Electric charge of an electron in Coulombs
"""


import json
import os

def magnetic_ion_j():
    #Loads j values for Magnetic ions.
    #Returns:  ion_j values : dict 
    with open(os.path.join(os.path.dirname(__file__),
                           "database/magnetic_form_factors.json"), 'r') as infile:
        return json.load(infile)

def periodic_table():
    #Loads periodic table database. mass, and long-form name.
    #Returns atom information dict, including mass, atomic number, density, mass, and name
    with open(os.path.join(os.path.dirname(__file__),
                           "database/periodic_table.json"), 'r') as infile:
        return json.load(infile)

def scattering_lengths():
    #Loads neutron scattering lengths.
    #Returns: scattering_lengths : dict

    with open(os.path.join(os.path.dirname(__file__),
                           "database/scattering_lengths.json"), 'r') as infile:
        return json.load(infile)

def symmetry():
    #Loads crystal lattice space groups.
    #Returns    lattice_space_groups : dict Database of 230 crystal lattice space groups and their generators
    with open(os.path.join(os.path.dirname(__file__),
                           "database/symmetry.json"), 'r') as infile:
        return json.load(infile)


JOULES_TO_MEV      = 1. / 1.6021766208e-19 * 1.e3  # Joules to meV
BOLTZMANN_IN_MEV_K = 8.6173303e-05 * 1.e3          # Boltzmann constant in meV/K
N_A                = 6.022140857e+23
neutron_mass       = 1.674927211e-24               # mass of a neutron in grams
hbar               = 1.054571628e-34               # hbar in m2 kg / s
e                  = 1.602176487e-19               # coulombs



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
# This code is adapted from neutronpy


#Chopper class for Time of Flight spectrometer


import numpy as np

from .exceptions import ChopperError


class Chopper(object):
    # Chopper Class for using Time of Flight spectrometer
 
    def __init__(self, distance, speed, width, chopper_type, acceptance, counter_rot=False, radius=None, depth=None, tau=None):
        self.distance = distance
        self.speed = speed
        self.width = width
        self.chopper_type = chopper_type
        self.acceptance = acceptance
        if counter_rot:
            self.counter_rot = 2.0
        else:
            self.counter_rot = 1.0
        if radius is not None:
            self.radius = radius
        if depth is not None:
            self.depth = depth
        if tau is not None:
            self.tau_override = tau


    def __repr__(self):
        args = ', '.join(
            [str(getattr(self, key)) for key in ['distance', 'speed', 'width', 'chopper_type', 'acceptance']])
        kwargs = ', '.join(
            ['{0}={1}'.format(getattr(self, key)) for key in ['depth', 'tau'] if getattr(self, key, None) is not None])
        return "Chopper({0})".format(', '.join([args, kwargs]))

    @property
    def tau(self):
        """Calculate the time resolution of the chopper

        Returns
        -------
        tau : float
            Returns the resolution of the chopper in standard deviation in units of microseconds
        """
        if hasattr(self, 'tau_override'):
            return self.tau_override

        elif self.chopper_type == 'disk' and hasattr(self, 'radius'):
            return self.acceptance / (self.radius * self.speed * self.counter_rot) / np.sqrt(8 * np.log(2))
        elif self.chopper_type == 'disk' and ~hasattr(self, 'radius'):
            return 1e6 / (self.speed * self.acceptance * self.counter_rot) / 360.0
        elif self.chopper_type == 'fermi':
            try:
                return 1e6 / (self.speed * 2.0 * np.arctan(self.acceptance / self.depth)) / 360.
            except AttributeError:
                raise ChopperError("'depth' not specified, and is a required value for a Fermi Chopper.")

        else:
            raise ChopperError("'{0}' is an invalid chopper_type. Choose 'disk' or 'fermi', or specify custom tau \
                                via `tau_override` attribute".format('chopper_type'))




class Detector(object):
    #Detector Class defining for Time of Flight spectrometer

    def __init__(self, shape, width, height, radius, hpixels, vpixels, tau=0.1, thickness=1, orientation=None, dead_angles=None):
        self.shape = shape
        self.width = width
        self.height = height
        self.radius = radius
        self.tau = tau
        self.thickness = thickness
        self.hpixels = hpixels
        self.vpixels = vpixels
        if dead_angles:
            self.dead_angles = dead_angles
        if orientation:
            self.orientation = orientation

    def __repr__(self):
        args = ', '.join([str(getattr(self, key)) for key in ['shape', 'width', 'height', 'radius']])
        kwargs = ', '.join(
            ['{0}={1}'.format(key, getattr(self, key, None)) for key in ['resolution', 'orientation', 'dead_angles']])
        return "Detector({0})".format(', '.join([args, kwargs]))
        

class Guide(object):
    #Neutron Guide Class

    def __init__(self, m, length, width, height):
        self.m = m
        self.length = length
        self.width = width
        self.height = height

    @property
    def sigma_l(self):
        return np.sqrt(self.length ** 2 + self.width ** 2 + self.height ** 2) - self.length

    @property
    def sigma_theta(self):
        return np.arcsin(np.sqrt(self.m * np.sin(2.03e-3) ** 2))

    @property
    def sigma_phi(self):
        return self.sigma_theta



#!/usr/bin/env python3no
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 30 13:27:29 2020

@author: dgc

Improved version using scipy.optimize for faster fitting
"""
import numpy as np
from numpy import matlib
import sys
import scipy as spy
from scipy.optimize import least_squares
import time

from ..instrument.tas_spectr import TripleAxisSpectr
from ..instrument.tools import _scalar, _star,_modvec, _cleanargs


class FitConv(object):
    """
    Improved FitConv class using scipy.optimize for faster fitting
    """
    
    def __init__(self, exp=None, sqw=None, prefactor=None, hkle=None, Iobs=None, dIobs=None, param=None, paramfixed=None, 
                 method=1, maxiter=200, tol=1e-6, dtol=1e-6):
        
        if exp is None:
            print("exp is not provided")
            exp=TripleAxisSpectr(efixed=5)
            return

        if sqw is None:
            print("sqw is not provided")
            return

        if prefactor is None:
            print("prefactor is not provided")
            return

        self.exp       =    exp
        self.sqw       =    sqw
        self.prefactor =    prefactor
        self.hkle      =    np.array(hkle)
        self.Iobs      =    Iobs
        self.dIobs     =    dIobs
        self.param     =    param
        self.paramfixed=    paramfixed
        self.method    =    method
        self.accuracy  =    None
        self.maxiter   =    maxiter
        self.tol       =    tol
        self.dtol      =    dtol
        self.fitcount  =    0
        self.chisq     =    0
        self.alpha     =    0
        self.beta      =    0
        

    def __repr__(self):
        return " this is a FitConv class for fitting the exp by convoluting the TAS instrument resolution"
    
    def __str__(self):
        return "this is a FitConv class"
    
    @property
    def exp(self):
        return self._exp

    @exp.setter
    def exp(self, value):
        self._exp = value

    @property
    def sqw(self):
        return self._sqw

    @sqw.setter
    def sqw(self, value):
        self._sqw = value  

    @property
    def prefactor(self):
        return self._prefactor

    @prefactor.setter
    def prefactor(self, value):
        self._prefactor = value  
        
    @property
    def hkle(self):
        return self._hkle

    @hkle.setter
    def hkle(self, value):
        self._hkle = value        
        
    @property
    def Iobs(self):
        return self._Iobs

    @Iobs.setter
    def Iobs(self, value):
        self._Iobs = value      
        

    @property
    def dIobs(self):
        return self._dIobs

    @dIobs.setter
    def dIobs(self, value):
        self._dIobs = value               

    @property
    def param(self):
        return self._param

    @param.setter
    def param(self, value):
        self._param = value       

    @property
    def paramfixed(self):
        return self._paramfixed

    @paramfixed.setter
    def paramfixed(self, value):
        self._paramfixed = value
        
    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        self._method = value            
        
    @property
    def maxiter(self):
        return self._maxiter

    @maxiter.setter
    def maxiter(self, value):
        self._maxiter = value    

    @property
    def tol(self):
        return self._tol

    @tol.setter  # Fixed: was @param.setter
    def tol(self, value):
        self._tol = value    
        
    @property
    def dtol(self):
        return self._dtol

    @dtol.setter
    def dtol(self, value):
        self._dtol = value            
    
    @property
    def fitcount(self):
        return self._fitcount

    @fitcount.setter
    def fitcount(self, value):
        self._fitcount = value

    @property
    def chisq(self):
        return self._chisq

    @chisq.setter
    def chisq(self, value):
        self._chisq = value    
        
    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        self._alpha = value    
        
    @property
    def beta(self):
        return self._beta

    @beta.setter
    def beta(self, value):
        self._beta = value    

    def residual_function(self, free_params):
        """
        Residual function for scipy.optimize.least_squares
        
        Parameters:
        free_params: array of free (varying) parameters
        
        Returns:
        residuals: weighted residuals for least squares fitting
        """
        # Reconstruct full parameter array
        full_params = self.param.copy()
        full_params[self.ivar] = free_params
        
        # Calculate model prediction
        f = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2, 
                           hkle=self.hkle, METHOD=self.method, 
                           ACCURACY=self.accuracy, p=full_params)
        self.fitcount += 1
        
        # Return weighted residuals
        return (self.Iobs - f) / self.dIobs

    def jacobian_function(self, free_params):
        """
        Jacobian function for scipy.optimize.least_squares
        More efficient than numerical differentiation
        
        Parameters:
        free_params: array of free (varying) parameters
        
        Returns:
        jacobian: matrix of partial derivatives
        """
        # Reconstruct full parameter array
        full_params = self.param.copy()
        full_params[self.ivar] = free_params
        
        npoint = len(self.Iobs)
        jacobian = np.zeros((npoint, len(free_params)))
        
        for i, param_idx in enumerate(self.ivar):
            # Calculate step size
            h = np.zeros(len(full_params))
            step = self.dtol * full_params[param_idx]
            if full_params[param_idx] == 0:
                step = 1e-8
            h[param_idx] = step
            
            # Forward and backward evaluations
            f1 = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                hkle=self.hkle, METHOD=self.method,
                                ACCURACY=self.accuracy, p=full_params + h)
            self.fitcount += 1
            
            f2 = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                hkle=self.hkle, METHOD=self.method,
                                ACCURACY=self.accuracy, p=full_params - h)
            self.fitcount += 1
            
            # Numerical derivative, normalized by error
            jacobian[:, i] = -(f1 - f2) / (2 * step * self.dIobs)
        
        return jacobian

    def fitwithconv(self, exp=None, sqw=None, prefactor=None, hkle=None, Iobs=None, dIobs=None, param=None, paramfixed=None, 
                 method='fix', maxiter=200, tol=1e-6, dtol=1e-6, use_jacobian=True):
        """
        Improved fitting function using scipy.optimize.least_squares
        
        Parameters:
        use_jacobian: bool, whether to use analytical jacobian (faster) or numerical (more robust)
        """
        
        t = time.time()
        self.exp = exp
        self.sqw = sqw
        self.prefactor = prefactor
        
        if hkle is not None and not isinstance(hkle, np.ndarray):
            self.hkle = np.array(hkle)

        if Iobs is not None and not isinstance(Iobs, np.ndarray):
            self.Iobs = np.array(Iobs)
            
        if dIobs is not None and not isinstance(dIobs, np.ndarray):
            self.dIobs = np.array(dIobs)
        
        if param is not None:
            self.param = np.array(param)  
           
        if paramfixed is not None:
            self.paramfixed = np.array(paramfixed)
        else:
            self.paramfixed = np.ones(self.param.shape)

        self.method = method
        self.accuracy = None
        self.maxiter = maxiter
        self.tol = tol
        self.dtol = dtol
        self.fitcount = 0
        
        # Process input data
        [H, K, L, W] = self.hkle
        dataLength = H.size

        if dIobs is None:
            dIobs = np.sqrt(self.Iobs)
        self.dIobs = dIobs.reshape(-1)

        if not dataLength == self.dIobs.size:
            print('Fatal error: size mismatch between H,K,L,W or EXP and dIobs')
            return
 
        if self.method is None:
            self.method = 'fix'
        
        if self.accuracy is None:
            self.accuracy = np.array([7, 0])

        # Set up optimization
        self.ivar = np.where(self.paramfixed > 0)[0]  # indices of varying parameters
        DF = len(self.Iobs) - len(self.ivar)  # degrees of freedom
        
        dispia = np.where(self.paramfixed > 0, 1, 0)
        print("Fitting {} to {} data points with {} free parameters and {} fixed parameters\n".format(
            sqw, len(self.Iobs), np.sum(dispia), len(self.param) - np.sum(dispia)))

        # Initial parameter values for free parameters only
        x0 = self.param[self.ivar]
        
        # Set up bounds (can be customized if needed)
        bounds = (-np.inf, np.inf)  # No bounds by default
        
        # Choose jacobian
        jac = self.jacobian_function if use_jacobian else '2-point'
        
        # Perform optimization using scipy's least_squares
        print('Using scipy.optimize.least_squares with trf method')
        print('Iteration info will be displayed by scipy...\n')
        
        # Callback function to display progress
        def callback(xk, rk=None):
            if hasattr(callback, 'nit'):
                callback.nit += 1
            else:
                callback.nit = 1
            
            # Calculate chi-squared for display
            residuals = self.residual_function(xk)
            chisq = np.sum(residuals**2)
            print(f' {callback.nit:8d} {self.fitcount:13d} {chisq/DF:13f} {(time.time()-t):13.1f}s')

        print('Iteration #   FunCount     chi^2/DF           time')
        
        # Run optimization
        result = least_squares(
            self.residual_function,
            x0,
            method='trf',  # Levenberg-Marquardt
            jac=jac,
            max_nfev=maxiter * len(x0) * 10,  # Maximum function evaluations
            xtol=tol,
            ftol=tol,
            verbose=0,  # We handle our own progress display
            callback=callback  # Uncomment if you want iteration display
        )
        
        # Update parameters
        self.param[self.ivar] = result.x
        
        # Calculate final statistics
        residuals = self.residual_function(result.x)
        chisq = np.sum(residuals**2)
        self.chisq = chisq
        
        # Calculate errors and correlation matrix
        # Use the Jacobian at the solution
        if result.jac is not None:
            J = result.jac
        else:
            J = self.jacobian_function(result.x)
        
        # Calculate covariance matrix
        try:
            # JTJ = J.T @ J, but we need to be careful about conditioning
            JTJ = np.dot(J.T, J)
            C = np.linalg.inv(JTJ)
        except np.linalg.LinAlgError:
            print("Warning: Singular matrix encountered. Using pseudo-inverse.")
            C = np.linalg.pinv(JTJ)
        
        # Parameter errors
        dpa = np.zeros(len(self.param))
        dpa[self.ivar] = np.sqrt(np.diag(C))
        
        # Normalized chi-squared
        chisqN = chisq / DF
        
        # Normalized correlation matrix
        CN = C / np.sqrt(np.abs(np.diag(C).reshape(-1, 1) * np.diag(C)))
        
        # Goodness of fit probability
        PQ = 1 - spy.special.gammainc(DF/2, chisq/2)
        
        # Final simulation
        sim = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                             hkle=self.hkle, METHOD=self.method,
                             ACCURACY=self.accuracy, p=self.param)
        self.fitcount += 1
        
        # Convergence analysis
        if result.success:
            kvg = 1
            print(f'\nStop: Converged normally in {result.nfev} function evaluations.')
        else:
            if result.nfev >= maxiter * len(x0) * 10:
                kvg = 0
                print('\nStop: max allowed number of function evaluations exceeded.')
            else:
                kvg = 2
                print(f'\nStop: {result.message}')
        
        print(f'\nTotal fitting time: {time.time() - t:.1f} seconds')
        print(f'Total function evaluations: {self.fitcount}')
        print(f'Final chi^2/DF: {chisqN:.6f}')
        print('\nFinal parameters:\n')
        
        for index, (par, err) in enumerate(zip(self.param, dpa)):
            fixed_str = "fixed" if self.paramfixed[index] == 0 else "free"
            print(f'{index:6d} \t {par:12.6f} ± {err:12.6f} ({fixed_str})')
        
        # Details dictionary
        details = {
            "chisq": chisq,
            "Ndata": len(self.Iobs),
            "Npar": len(self.param),
            "Nva": len(self.ivar),
            "DF": DF,
            "C": C,
            "result": result,
            "final_residuals": residuals
        }
        
        return [self.param, dpa, chisqN, sim, CN, PQ, result.nfev, kvg, details]

    # Keep the old methods for backward compatibility
    def dfdp(self, param, ivar):
        """Backward compatibility method - now uses the jacobian calculation"""
        [npoint, m] = self.hkle.shape
        dfp = np.zeros((npoint, len(ivar))) 
        
        for i, item in enumerate(ivar):
            h = np.zeros(len(param))
            t = param[item] + self.dtol * param[item]
            h[item] = t - param[item] 
            if param[item] == 0:
                h[item] = 1e-8
            
            pa1 = param + h
            pa2 = param - h
            
            f1 = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                hkle=self.hkle, METHOD=self.method,
                                ACCURACY=self.accuracy, p=pa1)  
            self.fitcount += 1
            
            f2 = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                hkle=self.hkle, METHOD=self.method,
                                ACCURACY=self.accuracy, p=pa2)
            self.fitcount += 1

            dfp[:, i] = (f1 - f2) / (2 * h[item])
            
        return dfp

    def marqit(self, param, ivar):
        """Backward compatibility method"""
        f = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                           hkle=self.hkle, METHOD=self.method,
                           ACCURACY=self.accuracy, p=param)
        self.fitcount += 1

        wdiff = (self.Iobs - f) / self.dIobs
        chisq = np.sum(wdiff**2)
                                        
        dfp = self.dfdp(param, ivar)
        NP = len(ivar)
        beta = np.sum(np.matlib.repmat(wdiff/self.dIobs, NP, 1) * dfp.T, 1)
        alpha = dfp.T / np.matlib.repmat(self.dIobs, NP, 1)
        alpha = np.matmul(alpha, alpha.T)
        
        return [chisq, alpha, beta]


# -*- coding: utf-8 -*-
# This code is adapted from reslib 4.2c


#Custom Exceptions for the instrument module

class DetectorError(Exception):
    #For Exceptions related to the properties of the Detector
    pass


class ScatteringTriangleError(Exception):
    #For Exceptions related to the Scattering Triangle not closing
    pass


class MonochromatorError(Exception):
    #For Exceptions related to the Monochromator properties
    pass


class AnalyzerError(Exception):
    #For Exceptions related to the Monochromator properties
    pass


class ChopperError(Exception):
    #For Exceptions related to the Chopper properties
    pass


class GoniometerError(Exception):
    #For Exceptions related to the Goniometer positions
    pass


class InstrumentError(Exception):
    #For general, unclassified types of Instrument Exceptions
    pass



from inspy.energy import Energy

from inspy.crystal import Sample
from inspy.instrument import get_tau, TripleAxisSpectr
from inspy.instrument import Mono
from inspy.instrument import Ana

import numpy as np




mono = Mono(tau='PG(002)', mosaic=30, vmosaic=30, height=20, width=20, depth=0.2, rh=150, rv=150, direct=-1)
ana = Ana(tau='PG(002)', mosaic=30, vmosaic=30, height=20, width=20, depth=0.2, rh=150, rv=150, direct=-1, thickness =0.2, horifoc=-1)
sample= Sample(3,4,5,90,90,90)

instr=TripleAxisSpectr(efixed=14.7, mono=mono, sample=sample, ana=ana)

hkle=np.array([[2, 0, 0, 0],[2, 0, 0, 2],[2, 0, 0, 2]])

print(hkle)

test=instr.CalcResMatHKL(hkle.T)

print(test)


# -*- coding: utf-8 -*-
# This file is adapted from neutronpy

# Material constructor

import inspy.constants as const
import numpy as np

from .atom import Atom, MagneticAtom
from .sample import Sample
from .structure_factors import MagneticStructureFactor, NuclearStructureFactor
from .symmetry import SpaceGroup

# from ..scattering.pattern import HKLGenerator


class MagneticUnitCell(Sample):
    # defining a magnetic unit cell

    def __init__(self, unit_cell):
        if 'chirality' not in unit_cell:
            self.chirality = 0
        if 'ph' not in unit_cell:
            self.phase = 0
        self.atoms = []
        if 'space_group' in unit_cell:
            self.space_group = SpaceGroup(unit_cell['space_group'])
            for atom in unit_cell['atoms']:
                symmetrized_positions = self.space_group.symmetrize_position(atom['pos'])
                for pos in symmetrized_positions:
                    self.atoms.append(MagneticAtom(atom['ion'],
                                                   pos,
                                                   atom['moment'],
                                                   atom['occupancy']))
        else:
            for atom in unit_cell['atoms']:
                self.atoms.append(MagneticAtom(atom['ion'],
                                               atom['pos'],
                                               atom['moment'],
                                               atom['occupancy']))

        self.propagation_vector = unit_cell['propagation_vector']

        a, b, c = unit_cell['lattice']['abc']
        alpha, beta, gamma = unit_cell['lattice']['abg']
        super(MagneticUnitCell, self).__init__(a, b, c, alpha, beta, gamma)

    def __repr__(self):
        return "MagneticUnitCell('{0}')".format(self.propagation_vector)


class Material(Sample, NuclearStructureFactor, MagneticStructureFactor):
    #Class for the Material being supplied for the structure factor calculation


    def __init__(self, crystal):
        self.name = crystal['name']

        if 'formulaUnits' not in crystal:
            crystal['formulaUnits'] = 1.

        self.muCell = 0.
        for item in crystal['composition']:
            if 'occupancy' not in item:
                item['occupancy'] = 1.
            self.muCell += const.periodic_table()[item['ion']]['mass'] * item['occupancy']

        self.Mcell = self.muCell * crystal['formulaUnits']

        if 'lattice' in crystal:
            a, b, c = crystal['lattice']['abc']
            alpha, beta, gamma = crystal['lattice']['abg']

        if 'wavelength' in crystal:
            self.wavelength = crystal['wavelength']
        else:
            self.wavelength = 2.359

        if 'space_group' in crystal:
            self.space_group = SpaceGroup(crystal['space_group'])
            self.atoms = []
            for atom in crystal['composition']:
                if 'Uiso' not in atom:
                    atom['Uiso'] = 0
                if 'Uaniso' not in item:
                    atom['Uaniso'] = np.matrix(np.zeros((3, 3)))
                if 'occupancy' not in item:
                    atom['occupancy'] = 1.
                symmetrized_positions = self.space_group.symmetrize_position(atom['pos'])
                for pos in symmetrized_positions:
                    self.atoms.append(Atom(atom['ion'],
                                           pos,
                                           atom['occupancy'],
                                           self.Mcell,
                                           crystal['massNorm'],
                                           atom['Uiso'],
                                           atom['Uaniso']))
        else:
            self.atoms = []
            for item in crystal['composition']:
                if 'Uiso' not in item:
                    item['Uiso'] = 0
                if 'Uaniso' not in item:
                    item['Uaniso'] = np.matrix(np.zeros((3, 3)))
                if 'occupancy' not in item:
                    item['occupancy'] = 1.
                self.atoms.append(Atom(item['ion'],
                                       item['pos'],
                                       item['occupancy'],
                                       self.Mcell,
                                       crystal['massNorm'],
                                       item['Uiso'],
                                       item['Uaniso']))

        if 'magnetic_unit_cell' in crystal:
            self.magnetic_unit_cell = MagneticUnitCell(crystal['magnetic_cell'])

        if 'mosaic' not in crystal:
            crystal['mosaic'] = None
        if 'vmosaic' not in crystal:
            crystal['vmosaic'] = None
        if 'u' not in crystal:
            crystal['u'] = None
        if 'v' not in crystal:
            crystal['v'] = None
        if 'dir' not in crystal:
            crystal['dir'] = 1

        super(Material, self).__init__(a, b, c, alpha, beta, gamma, crystal['mosaic'], crystal['vmosaic'],
                                       crystal['dir'], crystal['u'], crystal['v'])

    def __repr__(self):
        return "Material('{0}')".format(self.name)

    @property
    def total_scattering_cross_section(self):
        r"""Returns total scattering cross-section of unit cell
        """
        total = 0
        for atom in self.atoms:
            total += (atom.coh_xs + atom.inc_xs)
        return total

    def N_atoms(self, mass):
        #Number of atoms in the defined Material, given the mass of the sample.

        return const.N_A * mass / self.muCell

    def calc_optimal_thickness(self, energy=25.3, transmission=1 / np.exp(1)):
        #Calculates the optimal sample thickess to avoid problems with
        #extinction, multiple coherent scattering and absorption.


        sigma_coh = np.sum([atom.occupancy * atom.coh_xs for atom in self.atoms])
        sigma_inc = np.sum([atom.occupancy * atom.inc_xs for atom in self.atoms])
        sigma_abs = np.sum([atom.occupancy * atom.abs_xs for atom in self.atoms])

        sigma_T = (sigma_coh + sigma_inc + sigma_abs * np.sqrt(25.3 / energy)) / self.volume

        return -np.log(transmission) / sigma_T

    def calc_incoh_elas_xs(self, mass=None):
        #Calculates the incoherent elastic cross section.


        INC_XS = 0
        for atom in self.atoms:
            INC_XS += atom.inc_xs * np.exp(-8 * np.pi ** 2 * atom.Uiso * np.sin(
                np.deg2rad(self.get_two_theta(atom.pos, self.wavelength) / 2.)) ** 2 / self.wavelength ** 2)

        if mass is not None:
            return self.N_atoms(mass) / (4 * np.pi) * INC_XS
        else:
            return INC_XS / (4 * np.pi)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Ultra-Fast FitConv - Optimized for maximum speed
Removes interface compatibility constraints for performance gains
"""
import numpy as np
import scipy as spy
from scipy.optimize import least_squares
import time
from functools import lru_cache
import warnings

from ..instrument.tas_spectr import TripleAxisSpectr
from ..instrument.tools import _scalar, _star,_modvec, _cleanargs


class UltraFastFitConv(object):
    """
    Ultra-optimized FitConv class prioritizing speed over interface compatibility
    """
    
    def __init__(self, exp, sqw, prefactor, hkle, Iobs, dIobs=None, 
                 cache_size=128, precompute_data=True):
        """
        Streamlined initialization with mandatory parameters
        
        Parameters:
        exp: instrument object
        sqw: S(Q,w) model function  
        prefactor: prefactor function
        hkle: [H, K, L, E] arrays as numpy arrays
        Iobs: observed intensities as numpy array
        dIobs: intensity errors as numpy array (optional, will use sqrt(Iobs) if None)
        cache_size: LRU cache size for expensive calculations
        precompute_data: whether to precompute and validate data upfront
        """
        
        # Store references (no property overhead)
        self.exp = exp
        self.sqw = sqw
        self.prefactor = prefactor
        
        # Preprocess and validate data once
        if precompute_data:
            self._preprocess_data(hkle, Iobs, dIobs)
        else:
            self.hkle = np.asarray(hkle, dtype=np.float64)
            self.Iobs = np.asarray(Iobs, dtype=np.float64)
            self.dIobs = np.asarray(dIobs, dtype=np.float64) if dIobs is not None else np.sqrt(self.Iobs)
        
        # Performance tracking
        self.fitcount = 0
        self._cache_size = cache_size
        
        # Pre-allocate arrays for gradient calculations
        self._gradient_workspace = None
        self._last_params_hash = None
        
    def _preprocess_data(self, hkle, Iobs, dIobs):
        """Preprocess and validate all data once for maximum speed"""
        
        # Convert to optimal numpy arrays
        self.hkle = np.asarray(hkle, dtype=np.float64)
        self.Iobs = np.asarray(Iobs, dtype=np.float64).ravel()
        
        if dIobs is None:
            self.dIobs = np.sqrt(np.maximum(self.Iobs, 1e-10))  # Avoid division by zero
        else:
            self.dIobs = np.asarray(dIobs, dtype=np.float64).ravel()
            
        # Validate dimensions
        [H, K, L, W] = self.hkle
        self.npoints = H.size
        
        if self.Iobs.size != self.npoints:
            raise ValueError(f"Data size mismatch: hkle has {self.npoints} points, Iobs has {self.Iobs.size}")
        if self.dIobs.size != self.npoints:
            raise ValueError(f"Data size mismatch: hkle has {self.npoints} points, dIobs has {self.dIobs.size}")
            
        # Pre-calculate inverse errors for speed (avoid repeated division)
        self.inv_dIobs = 1.0 / self.dIobs
        
        # Check for problematic data
        if np.any(self.dIobs <= 0):
            warnings.warn("Zero or negative errors detected, replacing with sqrt(Iobs)")
            bad_errors = self.dIobs <= 0
            self.dIobs[bad_errors] = np.sqrt(np.maximum(self.Iobs[bad_errors], 1e-10))
            self.inv_dIobs = 1.0 / self.dIobs

    @lru_cache(maxsize=128)
    def _cached_model_eval(self, params_tuple, method='fix', accuracy_tuple=(7, 0)):
        """Cached model evaluation to avoid repeated identical calculations"""
        params_array = np.array(params_tuple)
        accuracy_array = np.array(accuracy_tuple)
        
        result = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                 hkle=self.hkle, METHOD=method,
                                 ACCURACY=accuracy_array, p=params_array)
        self.fitcount += 1
        return result

    def residual_vectorized(self, free_params, ivar, full_params_template, method='fix', accuracy=(7, 0)):
        """
        Ultra-fast vectorized residual function
        
        Parameters:
        free_params: array of free parameters only
        ivar: indices of free parameters
        full_params_template: template for full parameter array
        method: resolution method
        accuracy: accuracy parameters
        """
        # Reconstruct full parameters (fast in-place operation)
        full_params = full_params_template.copy()
        full_params[ivar] = free_params
        
        # Use caching for identical parameter sets
        params_tuple = tuple(full_params)
        accuracy_tuple = tuple(accuracy) if isinstance(accuracy, np.ndarray) else accuracy
        
        try:
            f = self._cached_model_eval(params_tuple, method, accuracy_tuple)
        except TypeError:
            # Fallback for non-hashable types
            f = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                               hkle=self.hkle, METHOD=method,
                               ACCURACY=np.array(accuracy), p=full_params)
            self.fitcount += 1
        
        # Vectorized residual calculation
        return (self.Iobs - f) * self.inv_dIobs

    def jacobian_optimized(self, free_params, ivar, full_params_template, 
                          method='fix', accuracy=(7, 0), rel_step=1e-8):
        """
        Optimized Jacobian calculation with adaptive step sizes
        """
        full_params = full_params_template.copy()
        full_params[ivar] = free_params
        
        nparams = len(free_params)
        jacobian = np.zeros((self.npoints, nparams))
        
        # Adaptive step size calculation
        steps = np.where(np.abs(free_params) > 1e-12, 
                        rel_step * np.abs(free_params),
                        rel_step)
        
        # Vectorized gradient computation where possible
        for i, param_idx in enumerate(ivar):
            step = steps[i]
            
            # Forward and backward parameter arrays
            params_forward = full_params.copy()
            params_backward = full_params.copy()
            params_forward[param_idx] += step
            params_backward[param_idx] -= step
            
            # Function evaluations with caching
            try:
                f_forward = self._cached_model_eval(tuple(params_forward), method, tuple(accuracy))
                f_backward = self._cached_model_eval(tuple(params_backward), method, tuple(accuracy))
            except TypeError:
                f_forward = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                           hkle=self.hkle, METHOD=method,
                                           ACCURACY=np.array(accuracy), p=params_forward)
                f_backward = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                            hkle=self.hkle, METHOD=method,
                                            ACCURACY=np.array(accuracy), p=params_backward)
                self.fitcount += 2
            
            # Central difference with pre-computed inverse errors
            jacobian[:, i] = -(f_forward - f_backward) * self.inv_dIobs / (2 * step)
        
        return jacobian

    def fit_ultrafast(self, param_initial, param_fixed_mask=None, 
                     method='fix', accuracy=(7, 0),
                     maxfev=None, xtol=1e-8, ftol=1e-8, gtol=1e-8,
                     use_analytical_jacobian=True, verbose=True,
                     early_stopping=True, convergence_window=5):
        """
        Ultra-fast fitting with all optimizations enabled
        
        Parameters:
        param_initial: initial parameter values
        param_fixed_mask: boolean mask (True = free, False = fixed)
        method: resolution convolution method
        accuracy: accuracy parameters
        maxfev: maximum function evaluations (auto-calculated if None)
        xtol, ftol, gtol: convergence tolerances
        use_analytical_jacobian: use our optimized jacobian
        verbose: display progress
        early_stopping: stop if convergence stalls
        convergence_window: window for early stopping detection
        """
        
        t_start = time.time()
        self.fitcount = 0
        
        # Parameter setup
        param_initial = np.asarray(param_initial, dtype=np.float64)
        if param_fixed_mask is None:
            param_fixed_mask = np.ones(len(param_initial), dtype=bool)
        else:
            param_fixed_mask = np.asarray(param_fixed_mask, dtype=bool)
        
        ivar = np.where(param_fixed_mask)[0]
        nfree = len(ivar)
        DF = self.npoints - nfree
        
        if maxfev is None:
            maxfev = max(200 * nfree, 1000)
        
        if verbose:
            print(f"Ultra-fast fitting: {self.npoints} data points, {nfree} free parameters")
            print(f"Method: {method}, Max evaluations: {maxfev}")
            print("FuncEvals     Chi²/DF        Time(s)")
            print("-" * 40)
        
        # Optimization setup
        x0 = param_initial[ivar]
        full_params_template = param_initial.copy()
        
        # Create optimized functions with pre-bound parameters
        def fast_residual(free_params):
            return self.residual_vectorized(free_params, ivar, full_params_template, method, accuracy)
        
        if use_analytical_jacobian:
            def fast_jacobian(free_params):
                return self.jacobian_optimized(free_params, ivar, full_params_template, method, accuracy)
            jac = fast_jacobian
        else:
            jac = '2-point'  # Let scipy handle it
        
        # Progress tracking for LM method
        progress_data = {'last_chi2': float('inf'), 'stall_count': 0, 'call_count': 0}
        
        def tracked_residual(free_params):
            progress_data['call_count'] += 1
            residuals = fast_residual(free_params)
            
            if verbose and progress_data['call_count'] % 20 == 0:  # Every 20 calls
                chi2 = np.sum(residuals**2) / DF
                elapsed = time.time() - t_start
                print(f"{self.fitcount:8d}   {chi2:10.6f}   {elapsed:8.2f}")
                
                # Early stopping check
                if early_stopping:
                    if abs(progress_data['last_chi2'] - chi2) < ftol * chi2:
                        progress_data['stall_count'] += 1
                    else:
                        progress_data['stall_count'] = 0
                    progress_data['last_chi2'] = chi2
                    
                    if progress_data['stall_count'] >= convergence_window:
                        if verbose:
                            print("Early stopping: convergence detected")
            
            return residuals
        
        # Run optimization with all speed optimizations
        try:
            result = least_squares(
                tracked_residual,
                x0,
                method='lm',
                jac=jac,
                max_nfev=maxfev,
                xtol=xtol,
                ftol=ftol,
                gtol=gtol,
                verbose=0
            )
        except KeyboardInterrupt:
            if verbose:
                print("\nFitting interrupted by user")
            raise
        
        # Process results
        final_params = full_params_template.copy()
        final_params[ivar] = result.x
        
        # Calculate final statistics
        final_residuals = fast_residual(result.x)
        chisq = np.sum(final_residuals**2)
        chisq_reduced = chisq / DF if DF > 0 else np.inf
        
        # Parameter errors from Jacobian
        param_errors = np.zeros(len(param_initial))
        if result.jac is not None:
            try:
                # Calculate covariance matrix
                JTJ = result.jac.T @ result.jac
                C = np.linalg.inv(JTJ)
                param_errors[ivar] = np.sqrt(np.diag(C))
            except np.linalg.LinAlgError:
                if verbose:
                    print("Warning: Could not calculate parameter errors (singular matrix)")
        
        # Final model evaluation
        final_model = self.exp.ResConv(sqw=self.sqw, pref=self.prefactor, nargout=2,
                                     hkle=self.hkle, METHOD=method,
                                     ACCURACY=np.array(accuracy), p=final_params)
        self.fitcount += 1
        
        elapsed_total = time.time() - t_start
        
        if verbose:
            print("-" * 40)
            print(f"Fitting completed in {elapsed_total:.2f} seconds")
            print(f"Total function evaluations: {self.fitcount}")
            print(f"Final χ²/DF: {chisq_reduced:.6f}")
            print(f"Convergence: {'Success' if result.success else 'Failed'}")
            print(f"Exit message: {result.message}")
            
            print("\nFinal Parameters:")
            for i, (val, err) in enumerate(zip(final_params, param_errors)):
                status = "free" if param_fixed_mask[i] else "fixed"
                print(f"  {i:2d}: {val:12.6f} ± {err:12.6f} ({status})")
        
        # Return comprehensive results
        return {
            'params': final_params,
            'param_errors': param_errors,
            'chi2_reduced': chisq_reduced,
            'chi2': chisq,
            'model': final_model,
            'residuals': final_residuals,
            'success': result.success,
            'message': result.message,
            'nfev': self.fitcount,
            'time_elapsed': elapsed_total,
            'degrees_of_freedom': DF,
            'scipy_result': result
        }

    def clear_cache(self):
        """Clear the LRU cache to free memory"""
        self._cached_model_eval.cache_clear()
        
    def get_cache_stats(self):
        """Get cache performance statistics"""
        return self._cached_model_eval.cache_info()


# -*- coding: utf-8 -*-

#Define a Time Of Flight instrument for resolution calculations
#unfinished

import numpy as np
from ..constants import e, hbar, neutron_mass
from ..crystal import Sample
from ..energy import Energy
from .components import Chopper, Detector,Guide
from .exceptions import DetectorError
from .tools import _cleanargs, chop, get_angle_ki_Q, get_kfree


class TimeOfFlightSpectr:
   
    

    def __init__(self, ei=3.0, choppers=None, sample=None, detector=None, guides=None, theta_i=0, phi_i=0, **kwargs):
        self._ei = Energy(energy=ei)
        return
        


    def __repr__(self):
        return "Instrument('tof',  ei={0})".format(self.ei)

    @property
    def ei(self):
        r"""Incident Energy object of type :py:class:`.Energy`

        """
        return self._ei

    @ei.setter
    def ei(self, value):
        self._ei = Energy(energy=value)

    @property
    def orient1(self):
        return self.sample.u

    @property
    def orient2(self):
        return self.sample.v

    def CalcResMatQ(self, Q, W):
        return


    def CalcResMatHKL(self, hkle):
        return



# -*- coding: utf-8 -*-
# This file is adapted from neutronpy

#Structure Factors


import numpy as np

from ..constants import magnetic_ion_j


class NuclearStructureFactor(object):
    #Class containing nuclear structure factor calculator


    def calc_nuc_str_fac(self, hkl):
        #Calculates the structural form factor of the material.
        h, k, l = hkl
        # Ensures input arrays are complex ndarrays
        if isinstance(h, (np.ndarray, list, tuple)):
            h = np.array(h).astype(complex)
        if isinstance(k, (np.ndarray, list, tuple)):
            k = np.array(k).astype(complex)
        if isinstance(l, (np.ndarray, list, tuple)):
            l = np.array(l).astype(complex)

        # construct structure factor
        NSF = 0 * 1j
        for atom in self.atoms:
            NSF += atom.occupancy * atom.b * np.exp(1j * 2. * np.pi * (h * atom.pos[0] + k * atom.pos[1] + l * atom.pos[2])) * \
                np.exp(-8 * np.pi ** 2 * atom.Uiso * np.sin(np.deg2rad(self.get_two_theta(atom.pos, self.wavelength) / 2.)) ** 2 / self.wavelength ** 2) * \
                np.exp(-np.float(np.dot(np.dot(atom.pos, atom.Uaniso), atom.pos)))

        return NSF


class MagneticFormFactor(object):
    #Class defining a magnetic ion.

    def __init__(self, ion):
        self.ion = ion
        try:
            self.j0 = magnetic_ion_j()[self.ion]['j0']
            self.j2 = magnetic_ion_j()[self.ion]['j2']
            self.j4 = magnetic_ion_j()[self.ion]['j4']
        except ValueError:
            raise ValueError('No such ion was found in database.')

    def __repr__(self):
        return "MagneticFormFactor('{0}')".format(self.ion)

    def calc_mag_form_fac(self, q=None, g=None, qrange=None):
        #Calculate the magnetic form factor of an ion.

        if q is None:
            if qrange is None:
                q = np.linspace(0., 2., 2. / 0.025 + 1)
            else:
                q = np.linspace(qrange[0], qrange[1], (qrange[1] - qrange[0]) / 0.025 + 1)
        if g is None:
            g = 2.

        x = q / 4. / np.pi

        j0 = (self.j0[0] * np.exp(-self.j0[1] * x ** 2) + self.j0[2] *
              np.exp(-self.j0[3] * x ** 2) + self.j0[4] *
              np.exp(-self.j0[5] * x ** 2) + self.j0[6])

        j2 = x ** 2 * (self.j2[0] * np.exp(-self.j2[1] * x ** 2) +
                       self.j2[2] * np.exp(-self.j2[3] * x ** 2) +
                       self.j2[4] * np.exp(-self.j2[5] * x ** 2) +
                       self.j2[6])

        j4 = x ** 2 * (self.j4[0] * np.exp(-self.j4[1] * x ** 2) +
                       self.j4[2] * np.exp(-self.j4[3] * x ** 2) +
                       self.j4[4] * np.exp(-self.j4[5] * x ** 2) +
                       self.j4[6])

        ff = j0 + (2. / g - 1.) * j2

        return ff, q, j0, j2, j4


class MagneticStructureFactor(object):
    #Class containing magnetic structure factor calculator

    def calc_mag_int_vec(self):
        #Calculates magnetic interaction vector
        pass

    def calc_mag_str_fac(self):
        #Calculates magnetic structure factor
        pass



# -*- coding: utf-8 -*-
# This code is adapted from reslib 4.2c

import numpy as np

from .tools import get_tau


"""A structure that describes the monochromator.

Attributes
----------
tau : str or float
    The monochromator reciprocal lattice vector in Å^(-1)
    One can use one of the following keyword strings or numerical numbers:

    +------------------+--------------+-----------+
    | String           |     τ        |           |
    +==================+==============+===========+
    | Be(002)          | 3.50702      |           |
    +------------------+--------------+-----------+
    | Co0.92Fe0.08(200)| 3.54782      | (Heusler) |
    +------------------+--------------+-----------+
    | Cu(002)          | 3.47714      |           |
    +------------------+--------------+-----------+
    | Cu(111)          | 2.99913      |           |
    +------------------+--------------+-----------+
    | Cu(220)          | 4.91642      |           |
    +------------------+--------------+-----------+
    | Cu2MnAl(111)     | 1.82810      | (Heusler) |
    +------------------+--------------+-----------+
    | Ge(111)          | 1.92366      |           |
    +------------------+--------------+-----------+
    | Ge(220)          | 3.14131      |           |
    +------------------+--------------+-----------+
    | Ge(311)          | 3.68351      |           |
    +------------------+--------------+-----------+
    | Ge(511)          | 5.76968      |           |
    +------------------+--------------+-----------+
    | Ge(533)          | 7.28063      |           |
    +------------------+--------------+-----------+
    | PG(002)          | 1.87325      |           |
    +------------------+--------------+-----------+
    | PG(004)          | 3.74650      |           |
    +------------------+--------------+-----------+
    | PG(110)          | 5.49806      |           |
    +------------------+--------------+-----------+
    | Si(111)          | 2.00421      |           |
    +------------------+--------------+-----------+

"""

class Mono(object):
    #Monochromator

    def __init__(self, tau, mosaic, vmosaic=None, height=None, width=None, depth=None, rh=None, rv=None, direct=-1):
        self._tau   = tau
        self.mosaic = mosaic
        if vmosaic is not None:
            self.vmosaic = vmosaic
        if rh is not None:
            self.rh = rh
        if rv is not None:
            self.rv = rv
        if height is not None:
            self.height = height
        if width  is not None:
            self.width  = width
        if depth  is not None:
            self.depth  = depth
        
        self.dir = direct
        self.d = 2 * np.pi / get_tau(tau)
        
        
    def __repr__(self):
        args   = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in ['tau', 'mosaic']])
        kwargs = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in
                            [ 'vmosaic', 'height', 'width', 'depth', 'rh', 'rv', 'direct'] if
                            getattr(self, key, None) is not None])
        return "Monochromator({0})".format(', '.join([args, kwargs]))

    def __eq__(self, right):
        self_parent_keys  = sorted(list(self.__dict__.keys()))
        right_parent_keys = sorted(list(right.__dict__.keys()))

        if not np.all(self_parent_keys == right_parent_keys):
            return False

        for key, value in self.__dict__.items():
            right_parent_val = getattr(right, key)
            if not np.all(value == right_parent_val):
                return False
        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    @property
    def tau(self):
        return self._tau

    @tau.setter
    def tau(self, tau):
        self._tau = tau
        self.d = 2 * np.pi / get_tau(tau)
        

    @property
    def direct(self):
        return self._dir

    @direct.setter
    def direct(self, value):
        self._dir = value
        
    @property
    def mosaic(self):
        return self._mosaic

    @mosaic.setter
    def mosaic(self, value):
        self._mosaic = value
        
    @property
    def vmosaic(self):
        return self._vmosaic

    @vmosaic.setter
    def vmosaic(self, value):
        self._vmosaic = value
    
    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        self._height = value

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        self._width = value
        
    @property
    def depth(self):
        return self._depth
    @depth.setter
    def depth(self, value):
        self._depth = value
        
    @property
    def rh(self):
        return self._rh
    @rh.setter
    def rh(self, value):
        self._rh = value  
        
    @property
    def rv(self):
        return self._rv
        
    @rv.setter
    def rv(self, value):
        self._rv = value        
        
        
        


# -*- coding: utf-8 -*-
# This file is adapted from neutronpy


import numpy as np
from ..constants import periodic_table, scattering_lengths


class Atom(object):
    #Class for adding atoms to the Material class.

    def __init__(self, ion, pos, occupancy=1., Mcell=None, massNorm=False, Uiso=0, Uaniso=np.zeros((3, 3))):
        self.ion = ion
        self.pos = np.array(pos)
        self.occupancy = occupancy
        self.Mcell = Mcell
        self.Uiso = Uiso
        self.Uaniso = np.matrix(Uaniso)

        if isinstance(scattering_lengths()[ion]['Coh b'], list):
            b = complex(*scattering_lengths()[ion]['Coh b'])
        else:
            b = scattering_lengths()[ion]['Coh b']

        if massNorm is True:
            self.mass = periodic_table()[ion]['mass']

            self.b = (b * self.occupancy * self.Mcell / np.sqrt(self.mass))
        else:
            self.b = b / 10.

        self.coh_xs = scattering_lengths()[ion]['Coh xs']
        self.inc_xs = scattering_lengths()[ion]['Inc xs']
        self.abs_xs = scattering_lengths()[ion]['Abs xs']

    def __repr__(self):
        return "Atom('{0}')".format(self.ion)


class MagneticAtom(object):
    #Class for adding magnetic atoms to the Material class.


    def __init__(self, ion, pos, moment, occupancy):
        self.ion = ion
        self.pos = np.array(pos)
        self.moment = moment
        self.occupancy = occupancy

    def __repr__(self):
        return "MagneticAtom('{0}')".format(self.ion, self.pos, self.moment, self.occupancy)
